@@ -1,35 +1,62 @@
 """
-The :class:`.BeamPath` is the main abstraction for the lightpath module,
-grouping together a set of devices using the :class:`.LightInterface` and
-representing the path between them as single object. While the manipulation of
-each of these object should be done at the device level, the
-:meth:`.BeamPath.clear` does provide a powerful tool to quickly change the
-status of the path.
+The :class:`.BeamPath` is the main abstraction for the Lightpath module,
+grouping together a set of devices using the :ref:`Lightpath
+interface<interface_api>` and representing the path between them as single
+object. The manipulation of each of these object should be done at the device
+level, and while this may be done inside of Lightpath via detailed screens,
+Lightpath is not responsible for inserting or removing devices.
 
 The :class:`.BeamPath` object is also not meant to be a rigid representation,
 :meth:`.BeamPath.split` and :meth:`.BeamPath.join` both allow for slicing and
 combining of different areas of the LCLS beamline. However, keep in mind that
 the path only knows the state of the devices it contains, so certain methods
 might not return an accurate representation of reality if an upstream device is
-affecting the beam.
+secretly affecting the beam.
 """
-import math
+from __future__ import annotations
+
 import enum
 import logging
-from collections import Iterable
+import math
+from collections import OrderedDict
+from collections.abc import Iterable
+from dataclasses import dataclass
+from typing import Callable, Dict, List, Optional, TextIO, Tuple, Union
 
-from prettytable import PrettyTable
+from ophyd import Device, DeviceStatus
 from ophyd.ophydobj import OphydObject
 from ophyd.status import wait as status_wait
 from ophyd.utils import DisconnectedError
+from prettytable import PrettyTable
 
-from .errors import CoordinateError
-
+from .errors import CoordinateError, PathError
 
 logger = logging.getLogger(__name__)
+# non-string sentinel for beginning of path
+NOT_A_DEVICE = 0
 
 
-class DeviceState(enum.Enum):
+@dataclass
+class LightpathState:
+    """
+    The lightpath-relevant state of a device. Devices must return an
+    instance of this dataclass via the ``get_lightpath_state()`` method.
+
+    Attributes
+    ----------
+    inserted : bool
+
+    removed : bool
+
+    output : Dict[str, float]
+        mapping from output branch to transmission
+    """
+    inserted: bool
+    removed: bool
+    output: Dict[str, float]
+
+
+class DeviceState(enum.IntEnum):
     """
     Description of BeamStates
 
@@ -68,46 +95,58 @@ class DeviceState(enum.Enum):
     Error = 5
 
 
-def find_device_state(device):
+def find_device_state(device: Device) -> Tuple[DeviceState, LightpathState]:
     """
     Report the state of a device
 
-    The device must implement ``.inserted`` and ``removed``.
+    The device must implement ``get_lightpath_state``, which returns a
+    ``LightpathState`` object.
 
     Parameters
     ----------
-    device : ophyd.Device
+    device : Device
+        ophyd Device implementing the Lightpath interface
 
     Returns
     -------
-    state: DeviceState
+    Tuple[DeviceState, LightpathState]
+        DeviceState enum
+        LightpathState dataclass
     """
     # Gather device information
     try:
-        _in, _out = device.inserted, device.removed
+        if not device.lightpath_summary.connected:
+            # Check if relevant signals are connected
+            logger.debug(f"Unable to connect to device: {device.name}")
+            return DeviceState.Disconnected, None
+
+        state = device.get_lightpath_state()
+        _in, _out = state.inserted, state.removed
         logger.debug("Device %s reporting; IN=%s, OUT=%s",
                      device.name, _in, _out)
-    # Check if this was an error with an EPICS connection
     except (TimeoutError, DisconnectedError) as exc:
+        # Technically it's possible to still have a timeout error, but
+        # it means our previous connection check gave a false positive
+        # severity should be a tad higher in this case
         logger.warning("Unable to connect to %r", device)
-        logger.debug(exc, exc_info=True)
-        return DeviceState.Disconnected
+        logger.warning(exc, exc_info=True)
+        return DeviceState.Disconnected, None
     except Exception:
         logger.exception("Unable to determine device state for %r", device)
-        return DeviceState.Error
+        return DeviceState.Error, None
     # Check state consistency and return proper Enum
     # In
     if _in and not _out:
-        return DeviceState.Inserted
+        return DeviceState.Inserted, state
     # Out
     elif _out and not _in:
-        return DeviceState.Removed
+        return DeviceState.Removed, state
     # Both In and Out
     elif _out and _in:
-        return DeviceState.Inconsistent
+        return DeviceState.Inconsistent, state
     # Neither In or Out
     else:
-        return DeviceState.Unknown
+        return DeviceState.Unknown, state
 
 
 class BeamPath(OphydObject):
@@ -122,7 +161,7 @@ class BeamPath(OphydObject):
     devices : :class:`.LightDevice`
         Arguments are interpreted as LightDevices along a common beamline.
 
-    name = str, optional
+    name : str, optional
         Name of the BeamPath
 
     Raises
@@ -144,15 +183,27 @@ class BeamPath(OphydObject):
     # Subscription Information
     SUB_PTH_CHNG = 'beampath_changed'
     _default_sub = SUB_PTH_CHNG
-    # Transmission setting
-    minimum_transmission = 0.1
 
-    def __init__(self, *devices, name=None):
+    def __init__(
+        self,
+        *devices: OphydObject,
+        minimum_transmission: float = 0.1,
+        name: Optional[str] = None,
+    ):
         super().__init__(name=name)
+        self.minimum_transmission = minimum_transmission
         self.devices = devices
         self._has_subscribed = False
+        self.branch_list = set()
         logger.debug("Configuring path %s with %s devices",
                      name, len(self.devices))
+
+        # create mapping of device to next device in z-order
+        sorted_devs = sorted(self.devices, key=lambda dev: dev.md.z)
+        self._next_device = OrderedDict({NOT_A_DEVICE: sorted_devs[0]})
+        self._next_device.update({sorted_devs[i].name: sorted_devs[i+1]
+                                  for i in range(len(sorted_devs) - 1)})
+
         # Sort by position downstream to upstream
         try:
             # Check types and positions
@@ -165,6 +216,9 @@ class BeamPath(OphydObject):
                                           'initialized', dev)
                 # Add as attribute
                 setattr(self, dev.name.replace(' ', '_'), dev)
+                # Update branch list
+                for br in dev.input_branches:
+                    self.branch_list.add(br)
 
         except AttributeError as e:
             raise TypeError('One of the devices does not meet the '
@@ -172,90 +226,126 @@ class BeamPath(OphydObject):
                             'attribute {}'.format(e))
 
     @property
-    def branches(self):
-        """
-        Branching devices along the path
-        """
-        return [d for d in self.devices if getattr(d, 'branches', False)]
+    def branching_devices(self) -> List[Device]:
+        """ List[Device]: Branching devices along the path """
+        return [d for d in self.devices
+                if len(getattr(d, 'output_branches', ['1'])) > 1]
 
     @property
-    def range(self):
-        """
-        Starting z position of beamline
-        """
+    def range(self) -> Tuple[float, float]:
+        """ Tuple[float, float]: Starting z position of beamline """
         return self.path[0].md.z, self.path[-1].md.z
 
     @property
-    def path(self):
+    def path(self) -> List[Device]:
+        """ List[Device]: List of devices ordered by coordinates """
+        return list(self._next_device.values())
+
+    def get_device_output(self, dev: Device) -> Tuple[str, float]:
         """
-        List of devices ordered by coordinates
+        Find relevant output item by attempting to match with the
+        input branch of the next device in this path.
+
+        Parameters
+        ----------
+        dev : Device
+            device in this path to get output from
         """
-        return sorted(self.devices, key=lambda dev: dev.md.z)
+        output = dev.get_lightpath_state().output
+
+        # get next device input branch
+        next_dev = self._next_device.get(dev.name, None)
+        if next_dev is None:
+            # last device has no successor
+            # take output on branch in branch list
+            output_keys = [br for br in output.keys()
+                           if br in self.branch_list]
+        else:
+            # base case, the device has a successor.
+            # match output with input of next device
+            next_in_brs = next_dev.input_branches
+            output_keys = [br for br in output.keys()
+                           if br in next_in_brs]
+
+        if len(output_keys) > 1:
+            raise PathError(f'Device {dev.name} has reported multiple '
+                            f'outputs along this path: {output_keys}')
+        elif len(output_keys) == 0:
+            return '', 0
+
+        return output_keys[0], output[output_keys[0]]
 
     @property
-    def blocking_devices(self):
+    def blocking_devices(self) -> List[Device]:
         """
         A list of devices that are currently inserted or are in unknown
         positions. This includes devices downstream of the first
-        :attr:`.impediment`
+        :attr:`.impediment`.
+
+        Returns
+        -------
+        List[Device]
+            list of blocking devices
         """
         # Cache important prior devices
-        prior = None
-        last_branches = list()
+        prev_device = None
+        prev_dev_branch = None
         block = list()
+        current_transmission = 1
         for device in self.path:
-            # If we have switched beamlines
-            if prior and device.md.beamline != prior.md.beamline:
-                # Find improperly configured optics
-                for optic in last_branches:
-                    # If this optic is responsible for delivering beam
-                    # to this hutch and it is not configured to do so.
-                    # Mark it as blocking
-                    if device.md.beamline in optic.branches:
-                        if device.md.beamline not in optic.destination:
-                            block.append(optic)
-                    # Otherwise ensure it is removed from the beamline
-                    elif optic.md.beamline not in optic.destination:
-                        block.append(optic)
-                # Clear optics that have been evaluated
-                last_branches.clear()
-
-            # If our last device was an optic, make sure it wasn't required
-            # to continue along this beampath
-            elif (prior in last_branches
-                    and device.md.beamline in prior.branches
-                    and device.md.beamline not in prior.destination):
-                block.append(last_branches.pop(-1))
-
-            # Find branching devices and store
-            # They will be marked as blocking by downstream devices
-            dev_state = find_device_state(device)
-            if device in self.branches:
-                last_branches.append(device)
-            # Find inserted devices
-            elif dev_state == DeviceState.Inserted:
-                # Ignore devices with low enough transmssion
-                trans = getattr(device, 'transmission', 1)
-                if trans < self.minimum_transmission:
-                    block.append(device)
-            # Find unknown and faulted devices
-            elif dev_state != DeviceState.Removed:
+            curr_state, curr_status = find_device_state(device)
+            # short circuit if statuses are in error
+            if curr_state in (DeviceState.Error, DeviceState.Unknown,
+                              DeviceState.Disconnected):
                 block.append(device)
-            # Stache our prior device
-            prior = device
+                continue
 
+            dev_out = self.get_device_output(device)
+            curr_dev_branch, curr_dev_trans = dev_out
+            # device output not on path
+            if curr_dev_branch == '':
+                block.append(device)
+            # check to make sure input and output branches match
+            # e.g. mirror not pointing to current device
+            elif (prev_device is not None and prev_dev_branch is not None
+                  and prev_dev_branch not in device.input_branches):
+                if prev_device not in block:
+                    block.append(prev_device)
+            # check inserted
+            elif curr_state is DeviceState.Inserted:
+                if curr_dev_trans > 1:
+                    logger.error(f'{device.name} reports transmission > 1')
+                current_transmission *= min(curr_dev_trans, 1)
+                # Any device seeing current transmission below min would block
+                if current_transmission < self.minimum_transmission:
+                    block.append(device)
+            # Do not add device to blocking list, it is removed
+            elif curr_state is DeviceState.Removed:
+                pass
+            # Unknown and inconsistent devices block
+            else:
+                block.append(device)
+
+            # stash previous device
+            prev_device = device
+            prev_dev_branch = curr_dev_branch
         return block
 
     @property
-    def incident_devices(self):
+    def incident_devices(self) -> List[Device]:
         """
         A list of devices the beam is currently incident on. This includes the
         current :attr:`.impediment` and any upstream devices that may be
         inserted but have more transmission than :attr:`.minimum_transmission`
+
+        Returns
+        -------
+        List[Device]
+            List of incident devices
         """
         # Find device information
         inserted = [d for d in self.path
-                    if find_device_state(d) == DeviceState.Inserted]
+                    if find_device_state(d)[0] == DeviceState.Inserted]
         impediment = self.impediment
         # No blocking devices, all inserted devices incident
         if not impediment:
@@ -263,17 +353,19 @@ class BeamPath(OphydObject):
         # Otherwise only return upstream of the impediment
         return [d for d in inserted if d.md.z <= impediment.md.z]
 
-    def show_devices(self, file=None):
+    def show_devices(self, file: TextIO = None):
         """
         Print a table of the devices along the beamline
 
         Parameters
         ----------
-        file : file-like object
-            File to writable
+        file : TextIO
+            File-like object to write output to.  Default behavior is
+            printing to sys.stdout
         """
         # Initialize Table
-        pt = PrettyTable(['Name', 'Prefix', 'Position', 'Beamline', 'State'])
+        pt = PrettyTable(['Name', 'Prefix', 'Position', 'Input Branches',
+                          'Output Branches', 'State'])
         # Adjust Table settings
         pt.align = 'r'
         pt.align['Name'] = 'l'
@@ -281,16 +373,14 @@ class BeamPath(OphydObject):
         pt.float_format = '8.5'
         # Add info
         for d in self.path:
-            pt.add_row([d.name, d.prefix, d.md.z,
-                        d.md.beamline, find_device_state(d).name])
+            pt.add_row([d.name, d.prefix, d.md.z, d.input_branches,
+                        d.output_branches, find_device_state(d)[0].name])
         # Show table
         print(pt, file=file)
 
     @property
-    def impediment(self):
-        """
-        First blocking device along the path
-        """
+    def impediment(self) -> Device:
+        """ Device: First blocking device along the path """
         # Find device information
         blocks = self.blocking_devices
         if not blocks:
@@ -300,21 +390,31 @@ class BeamPath(OphydObject):
             return blocks[0]
 
     @property
-    def cleared(self):
+    def cleared(self) -> bool:
         """
         Whether beamline is clear of any devices that are below the
         :attr:`.minimum_transmission`
+
+        Returns
+        -------
+        bool
+            whether beamline is clear of impediments
         """
         return not any(self.blocking_devices)
 
-    def clear(self, wait=False, timeout=None,
-              passive=False, ignore=None):
+    def clear(
+        self,
+        wait: bool = False,
+        timeout: Optional[float] = None,
+        ignore: Optional[List[Device]] = None,
+        passive: bool = False,
+    ) -> List[DeviceStatus]:
         """
         Clear the beampath of all obstructions
 
         Parameters
         ----------
-        wait : bool
+        wait : bool, optional
             Wait for all devices to complete their motion
 
         timeout : float, optional
@@ -329,7 +429,7 @@ class BeamPath(OphydObject):
 
         Returns
         -------
-        statuses :
+        statuses : List[ophyd.DeviceStatus]
             Returns list of status objects returned by
             :meth:`.LightInterface.remove`
         """
@@ -340,8 +440,8 @@ class BeamPath(OphydObject):
         logger.info('Removing devices along the beampath ...')
         status = [device.remove(timeout=timeout)
                   for device in target_devices
-                  if find_device_state(device) in (DeviceState.Inserted,
-                                                   DeviceState.Unknown)
+                  if find_device_state(device)[0] in (DeviceState.Inserted,
+                                                      DeviceState.Unknown)
                   and hasattr(device, 'remove')]
         # Wait parameters
         if wait:
@@ -356,7 +456,7 @@ class BeamPath(OphydObject):
 
         return status
 
-    def join(self, *beampaths):
+    def join(self, *beampaths: BeamPath) -> BeamPath:
         """
         Join multiple beampaths with the current one
 
@@ -378,7 +478,11 @@ class BeamPath(OphydObject):
         """
         return BeamPath.from_join(self, *beampaths, name=self.name)
 
-    def split(self, z=None, device=None):
+    def split(
+        self,
+        z: Optional[float] = None,
+        device: Optional[Device] = None
+    ) -> Tuple[BeamPath, BeamPath]:
         """
         Split the beampath producing two new BeamPath objects either by a
         specific position or a devices location
@@ -388,7 +492,7 @@ class BeamPath(OphydObject):
         z : float
             Z position to split the paths
 
-        device  : LightDevice, name, or base PV
+        device : LightDevice, name, or base PV
             The specified device will be the last device in the first
             :class:`.BeamPath` object
 
@@ -399,7 +503,9 @@ class BeamPath(OphydObject):
         """
         # Not enough information
         if not z and not device:
-            raise ValueError("Must supply information where to split the path")
+            raise ValueError(
+                "Must supply where to split the path (either z or device)"
+            )
         # Grab the z if given a device
         if device:
             z = device.md.z
@@ -408,11 +514,15 @@ class BeamPath(OphydObject):
             raise ValueError("Split position {} is not within the range of "
                              "the path.".format(z))
         # Split the paths
-        return (BeamPath(*[d for d in self.devices if d.md.z <= z]),
-                BeamPath(*[d for d in self.devices if d.md.z > z]))
+        return (BeamPath(*[d for d in self.devices if d.md.z <= z],
+                         name=f'{self.name}_pre_{z}',
+                         minimum_transmission=self.minimum_transmission),
+                BeamPath(*[d for d in self.devices if d.md.z > z],
+                         name=f'{self.name}_pre_{z}',
+                         minimum_transmission=self.minimum_transmission))
 
     @classmethod
-    def from_join(cls, *beampaths, name=None):
+    def from_join(cls, *beampaths: BeamPath, name: str = None) -> BeamPath:
         """
         Join other beampaths with the current one
 
@@ -443,14 +553,18 @@ class BeamPath(OphydObject):
         # Create a new instance
         return BeamPath(*set(devices), name=name)
 
-    def _ignore(self, ignore_devices, passive=False):
+    def _ignore(
+        self,
+        ignore_devices: Optional[Union[Device, List[Device]]] = None,
+        passive: bool = False
+    ) -> Tuple[List[Device], List[Device]]:
         """
         Assemble list of available devices with some exclusions
 
         Parameters
         ----------
         ignore_devices : list
-            Devices to ignore
+            Device(s) to ignore
 
         passive : bool
             If False, ignore passive devices
@@ -465,8 +579,10 @@ class BeamPath(OphydObject):
         # Add passive devices to ignored
         if not passive:
             logger.debug("Passive devices will be ignored ...")
-            ignore.extend([d for d in self.devices
-                           if d.transmission > self.minimum_transmission])
+            for dev in self.devices:
+                _, trans = self.get_device_output(dev)
+                if trans > self.minimum_transmission:
+                    ignore.append(dev)
         # Add ignored devices
         if isinstance(ignore_devices, Iterable):
             ignore.extend(ignore_devices)
@@ -490,10 +606,15 @@ class BeamPath(OphydObject):
         else:
             block = math.inf
         # If device is upstream of impediment
-        if obj and obj.md.z <= block:
+        if obj is not None and obj.parent.md.z <= block:
             self._run_subs(sub_type=self.SUB_PTH_CHNG, device=obj)
 
-    def subscribe(self, cb, event_type=None, run=True):
+    def subscribe(
+        self,
+        cb: Callable,
+        event_type: Optional[str] = None,
+        run: bool = True
+    ):
         """
         Subscribe to changes of the valve
 
@@ -513,15 +634,25 @@ class BeamPath(OphydObject):
             for dev in self.devices:
                 # Add callback here!
                 try:
-                    dev.subscribe(self._device_moved,
-                                  event_type=dev.SUB_STATE,
-                                  run=False)
+                    dev.lightpath_summary.subscribe(self._device_moved,
+                                                    run=False)
                 except Exception:
                     logger.error("BeamPath is unable to subscribe "
                                  "to device %s", dev.name)
             self._has_subscribed = True
         super().subscribe(cb, event_type=event_type, run=run)
 
+    def clear_device_subs(self) -> None:
+        """
+        Clears the ._device_moved callbacks from all devices in the path.
+        Distinct from BeamPath.clear_sub, which unsubscribes a specific
+        callback from the BeamPath object itself.
+        """
+        if self._has_subscribed:
+            for dev in self.devices:
+                dev.lightpath_summary.clear_sub(self._device_moved)
+            self._has_subscribed = False
+
     def _repr_info(self):
         yield('range',   self.range)
         yield('devices', len(self.devices))
